"""
Tests for Django Page Templates

$Id: test_djangopagetemplate.py,v 1.11 2005/11/20 10:40:19 stefan Exp $
"""

from unittest import TestCase, TestSuite, makeSuite, main

from os import environ

if not environ.get('DJANGO_SETTINGS_MODULE'):
    from django.contrib.pagetemplate.tests import settings
    environ['DJANGO_SETTINGS_MODULE'] = settings.__name__

from os import curdir
from os.path import realpath, dirname, join

here = realpath(dirname(__file__) or curdir)
layer1 = join(here, 'layer1')
layer2 = join(here, 'layer2')

from django.contrib.pagetemplate import pagetemplate
from django.contrib.pagetemplate.pagetemplate import *


class TemplateDirs(object):
    _dirs = ()
    def __init__(self):
        self._dirs = pagetemplate.TEMPLATE_DIRS
        pagetemplate.TEMPLATE_DIRS = (layer1, layer2)
    def restore(self):
        pagetemplate.TEMPLATE_DIRS = self._dirs


class Context(object):
    def __init__(self, dict=None):
        self.dicts = []
        if dict is not None:
            self.push(dict)
    def push(self, dict):
        self.dicts = [dict] + self.dicts


class TemplateTests(TestCase):

    def setUp(self):
        self.dirs = TemplateDirs()
        pagetemplate._cache.clear()

    def tearDown(self):
        self.dirs.restore()

    def test_findtemplate(self):
        # Should find templates in all layers
        p = findtemplate('main')
        self.assertEqual(p, join(layer1, 'main.pt'))
        p = findtemplate('main.pt')
        self.assertEqual(p, join(layer1, 'main.pt'))
        p = findtemplate('simple')
        self.assertEqual(p, join(layer2, 'simple.pt'))
        p = findtemplate('simple.pt')
        self.assertEqual(p, join(layer2, 'simple.pt'))

    def test_findtemplate_in_subdir(self):
        # Should find templates in subdirectories
        p = findtemplate('subdir1/template1')
        self.assertEqual(p, join(layer1, 'subdir1', 'template1.pt'))
        p = findtemplate('subdir2/template2')
        self.assertEqual(p, join(layer2, 'subdir2', 'template2.pt'))

    def test_missing_template(self):
        # A missing template should raise TemplateDoesNotExist
        self.assertRaises(TemplateDoesNotExist, findtemplate, 'unknown')

    def test_empty_template_dirs(self):
        # Empty TEMPLATE_DIRS should raise TemplateDoesNotExist
        pagetemplate.TEMPLATE_DIRS = ()
        self.assertRaises(TemplateDoesNotExist, findtemplate, 'main')

    def test_instantiate_template(self):
        # Should be able to create template objects
        t = DjangoPageTemplate('main')
        self.failUnless('main template' in t())
        t = DjangoPageTemplate('subdir1/template1')
        self.failUnless('template1' in t())

    def test_factory_method(self):
        # Factory method should return templates
        t = get_template('main')
        self.failUnless('main template' in t())

    def test_template_cache(self):
        # Factory method should add templates to cache
        t = get_template('main')
        self.failUnless('main' in pagetemplate._cache)

    def test_render(self):
        # Templates should be renderable
        t = get_template('main')
        c = Context()
        r = t.render(c)
        self.failUnless('main template' in r)
        self.failUnless('default title' in r)

    def test_namespace_from_context(self):
        # Values from the Context should be in the namespace
        t = get_template('main')
        c = Context({'title': 'context title'})
        self.failUnless('context title' in t.render(c))

    def test_context_priority(self):
        # Values "in front" should override values "in back"
        t = get_template('main')
        c = Context({'title': 'context title'})
        c.push({'title': 'alternative title'})
        self.failUnless('alternative title' in t.render(c))

    def test_namespace_from_kwargs(self):
        # Values passed as keyword arguments should be in the namespace
        t = get_template('main')
        c = Context()
        self.failUnless('kw title' in t.render(c, title='kw title'))

    def test_kwargs_priority(self):
        # Context values should override keyword arguments
        t = get_template('main')
        c = Context({'title': 'context title'})
        self.failUnless('context title' in t.render(c, title='kw title'))

    def test_namespace_from_defaults(self):
        # Default PT names should be in the namespace
        t = get_template('main')
        c = Context()
        self.failUnless('options: {}' in t.render(c))

    def test_defaults_priority(self):
        # Default names should override both context values and keyword
        # arguments
        t = get_template('main')
        c = Context({'options': {'fred': 'barney'}})
        self.failUnless('options: {}' in t.render(c))
        self.failUnless('options: {}' in t.render(c, options={'wilma': 'betty'}))

    def test_loadmacro(self):
        # Should be able to load METAL macros from templates
        m = loadmacro('main', 'master')
        # A "macro" is a list of stuff...
        self.failUnless(isinstance(m, list))
        self.assertEqual(m[0][0], 'version')

    def test_missing_macro(self):
        # A missing macro should raise MacroDoesNotExist
        self.assertRaises(MacroDoesNotExist, loadmacro, 'main', 'unknown')

    def test_broken_template(self):
        # A broken template should raise MacroDoesNotExist
        from logging import root
        root.disabled = True # Avoid log spam
        try:
            self.assertRaises(MacroDoesNotExist, loadmacro, 'broken', 'unknown')
        finally:
            root.disabled = False

    def test_broken_macro(self):
        # A broken macro should raise MacroDoesNotExist
        from logging import root
        root.disabled = True # Avoid log spam
        try:
            self.assertRaises(MacroDoesNotExist, loadmacro, 'bad', 'content')
        finally:
            root.disabled = False

    def test_macroless_template_explanation(self):
        # A macroless template has an empty macros dict
        t = get_template('simple')
        self.assertEqual(t.macros, {})

    def test_broken_template_explanation(self):
        # If the template is broken, the macros attribute is None!
        from logging import root
        root.disabled = True # Avoid log spam
        try:
            t = get_template('broken')
            self.assertEqual(t.macros, None)
        finally:
            root.disabled = False

    def test_broken_macro_explanation(self):
        # If the macro is broken, the macros attribute is None!
        from logging import root
        root.disabled = True # Avoid log spam
        try:
            t = get_template('bad')
            self.assertEqual(t.macros, None)
        finally:
            root.disabled = False

    def test_unicode(self):
        # Page Templates return unicode!
        t = get_template('main')
        c = Context()
        self.failUnless(isinstance(t.render(c), unicode))

    def test_latin1_umlauts(self):
        # Non-ASCII characters raise UnicodeDecodeError
        t = get_template('main')
        c = Context({'title': 'M\xe4dchen'}) # latin-1
        self.assertRaises(UnicodeDecodeError, t.render, c)

    def test_unicode_umlauts(self):
        # Non-ASCII characters must be passed as unicode
        t = get_template('main')
        c = Context({'title': unicode('M\xe4dchen', 'latin-1')})
        r = t.render(c)
        self.failUnless(isinstance(r, unicode))
        self.failUnless(u'M\xe4dchen' in r)

    def test_implicit_conversion(self):
        # Implicit charset conversion works if the default encoding
        # is set appropriately
        t = get_template('main')
        c = Context({'title': 'M\xe4dchen'}) # latin-1
        import sys
        reload(sys) # Get setdefaultencoding back
        default = sys.getdefaultencoding()
        sys.setdefaultencoding('latin-1')
        try:
            r = t.render(c)
        finally:
            sys.setdefaultencoding(default)
        self.failUnless(isinstance(r, unicode))
        self.failUnless(u'M\xe4dchen' in r)


class MacroLoaderTests(TestCase):

    def setUp(self):
        self.dirs = TemplateDirs()
        pagetemplate._cache.clear()

    def tearDown(self):
        self.dirs.restore()

    def test_loadmacro(self):
        # MacroLoader should load macros
        m = MacroLoader()['main']['macros']['master']
        # A "macro" is a list of stuff...
        self.failUnless(isinstance(m, list))
        self.assertEqual(m[0][0], 'version')

    def test_loadmacro_in_subdir(self):
        # MacroLoader should handle subdirectories
        m = MacroLoader()['subdir1']['template1']['macros']['content']
        # A "macro" is a list of stuff...
        self.failUnless(isinstance(m, list))
        self.assertEqual(m[0][0], 'version')

    def test_missing_macro(self):
        # A missing macro should raise MacroDoesNotExist
        try:
            MacroLoader()['main']['macros']['unknown']
        except MacroDoesNotExist:
            pass # test passed
        else:
            self.fail()

    def test_loader_error_1(self):
        # The MacroLoader should raise MacroLoaderError if the
        # path expression does not result in a macro spec.
        try:
            MacroLoader()()
        except MacroLoaderError:
            pass
        else:
            self.fail()

    def test_loader_error_2(self):
        # The MacroLoader should raise MacroLoaderError if the
        # path expression does not result in a macro spec.
        try:
            MacroLoader()['main']()
        except MacroLoaderError:
            pass
        else:
            self.fail()

    def test_loader_error_3(self):
        # The MacroLoader should raise MacroLoaderError if the
        # path expression does not result in a macro spec.
        try:
            MacroLoader()['macros']()
        except MacroLoaderError:
            pass
        else:
            self.fail()

    def test_loader_error_4(self):
        # The MacroLoader should raise MacroLoaderError if the
        # path expression does not result in a macro spec.
        try:
            MacroLoader()['main']['macros']()
        except MacroLoaderError:
            pass
        else:
            self.fail()


from django.contrib.pagetemplate.engine import simpleTraverse

class AttrTraversable(object):
    attr = 'foo'

class ItemTraversable(object):
    def __getitem__(self, name):
        if name == 'attr':
            return 'foo'
        raise KeyError, name

class AllTraversable(AttrTraversable, ItemTraversable):
    pass

_marker = object()

def getitem(ob, name, default=_marker):
    try:
        item = ob[name]
    except KeyError:
        if default is not _marker:
            return default
        raise KeyError, name
    else:
        return item


class PathExprTests(TestCase):

    def setUp(self):
        self.dirs = TemplateDirs()
        pagetemplate._cache.clear()

    def tearDown(self):
        self.dirs.restore()

    def testGetItem(self):
        # getitem helper should behave like __getitem__
        ob = {'attr': 'foo'}
        self.assertEqual(getitem(ob, 'attr', None), 'foo')
        self.assertEqual(getitem(ob, 'attr'), 'foo')
        self.assertEqual(getitem(ob, 'missing_attr', None), None)
        self.assertRaises(KeyError, getitem, ob, 'missing_attr')
        self.assertRaises(TypeError, getitem, object(), 'attr')

    def testAttrTraversable(self):
        # An object without __getitem__ should raise AttributeError
        # for missing attribute.
        ob = AttrTraversable()
        self.assertEqual(getattr(ob, 'attr', None), 'foo')
        self.assertRaises(AttributeError, getattr, ob, 'missing_attr')

    def testItemTraversable(self):
        # An object with __getitem__ (but without attr) should raise
        # KeyError for missing attribute.
        ob = ItemTraversable()
        self.assertEqual(getitem(ob, 'attr', None), 'foo')
        self.assertRaises(KeyError, getitem, ob, 'missing_attr')

    def testAllTraversable(self):
        # An object with attr and __getitem__ should raise either
        # exception, depending on method of access.
        ob = AllTraversable()
        self.assertEqual(getattr(ob, 'attr', None), 'foo')
        self.assertRaises(AttributeError, getattr, ob, 'missing_attr')
        self.assertEqual(getitem(ob, 'attr', None), 'foo')
        self.assertRaises(KeyError, getitem, ob, 'missing_attr')

    def testTraverseEmptyPath(self):
        # simpleTraverse should return the original object if the path is emtpy
        ob = object()
        self.assertEqual(simpleTraverse(ob, [], None), ob)

    def testTraverseByAttr(self):
        # simpleTraverse should find attr through attribute access
        ob = AttrTraversable()
        self.assertEqual(simpleTraverse(ob, ['attr'], None), 'foo')

    def testTraverseByMissingAttr(self):
        # simpleTraverse should raise AttributeError
        ob = AttrTraversable()
        # Here lurks the bug
        self.assertRaises(AttributeError, simpleTraverse, ob, ['missing_attr'], None)

    def testTraverseByItem(self):
        # simpleTraverse should find attr through item access
        ob = ItemTraversable()
        self.assertEqual(simpleTraverse(ob, ['attr'], None), 'foo')

    def testTraverseByMissingItem(self):
        # simpleTraverse should raise KeyError
        ob = ItemTraversable()
        self.assertRaises(KeyError, simpleTraverse, ob, ['missing_attr'], None)

    def testTraverseByAll(self):
        # simpleTraverse should find attr through attribute access
        ob = AllTraversable()
        self.assertEqual(simpleTraverse(ob, ['attr'], None), 'foo')

    def testTraverseByMissingAll(self):
        # simpleTraverse should raise KeyError (because ob implements __getitem__)
        ob = AllTraversable()
        self.assertRaises(KeyError, simpleTraverse, ob, ['missing_attr'], None)

    # pathexpr: <div tal:condition="object/?attr|nothing">foo</div>

    def testPathExprByAttr(self):
        # Condition should be True
        t = get_template('pathexpr')
        c = Context({'object': AttrTraversable(), 'attr': 'attr'})
        r = t.render(c)
        self.failUnless('foo' in r)

    def testPathExprByMissingAttr(self):
        # Condition should be False (nothing)
        t = get_template('pathexpr')
        c = Context({'object': AttrTraversable(), 'attr': 'missing_attr'})
        # Here lurks the bug
        r = t.render(c)
        self.failIf('foo' in r)

    def testPathExprByItem(self):
        # Condition should be True
        t = get_template('pathexpr')
        c = Context({'object': ItemTraversable(), 'attr': 'attr'})
        r = t.render(c)
        self.failUnless('foo' in r)

    def testPathExprByMissingItem(self):
        # Condition should be False (nothing)
        t = get_template('pathexpr')
        c = Context({'object': ItemTraversable(), 'attr': 'missing_attr'})
        r = t.render(c)
        self.failIf('foo' in r)

    def testPathExprByAll(self):
        # Condition should be True
        t = get_template('pathexpr')
        c = Context({'object': AllTraversable(), 'attr': 'attr'})
        r = t.render(c)
        self.failUnless('foo' in r)

    def testPathExprByMissingAll(self):
        # Condition should be False (nothing)
        t = get_template('pathexpr')
        c = Context({'object': AllTraversable(), 'attr': 'missing_attr'})
        r = t.render(c)
        self.failIf('foo' in r)


def test_suite():
    return TestSuite((
        makeSuite(TemplateTests),
        makeSuite(MacroLoaderTests),
        makeSuite(PathExprTests),
    ))


if __name__ == '__main__':
    main(defaultTest='test_suite')

